package com.miya.huihua.microservicebase.mq;

import com.aliyun.openservices.ons.api.*;
import com.miya.huihua.commons.exceptions.HuihuaException;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.SmartLifecycle;

import java.util.List;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

import static com.miya.huihua.microservicebase.common.Constants.ONS_MAX_RETRY_TIMES;
import static com.miya.huihua.microservicebase.common.Constants.TAG_ALL;
import static com.miya.huihua.microservicebase.common.ResultCode.INVALID_PARAMS;

/**
 * @author zhouxu
 * @createTime 2020/01/19
 */
@Slf4j
public abstract class AliConsumer implements SmartLifecycle {

    @Autowired(required = false)
    private OnsManager onsManager;

    /**
     * 订阅消息的topic
     *
     * @return
     */
    public abstract String getTopic();

    /**
     * 订阅消息的tag
     *
     * @return
     */
    public abstract List<String> getTags();

    public volatile boolean isRunning = false;

    /**
     * 消息处理
     *
     * @param tag
     * @param message
     */
    public abstract void consumerMessage(String tag, String message);

    @Override
    public boolean isAutoStartup() {
        return true;
    }

    @Override
    public void stop(Runnable runnable) {
        runnable.run();
        isRunning = false;
    }

    @Override
    public void start() {
        Consumer consumer = getConsumer(getTopic());
        if (!consumer.isStarted()) {
            //开始订阅消息
            consumer.subscribe(getTopic(), getTags().isEmpty() ? TAG_ALL : getTags().stream().collect(Collectors.joining("||")),
                    new OnsMessageListener((tag, messageBody) -> {
                        consumerMessage(tag, messageBody);
                    }));
            consumer.start();
            log.info("订阅topic：{},tag:{} 的消息队列已启动",getTopic(),getTags().isEmpty() ? TAG_ALL : getTags().stream().collect(Collectors.joining("||")));
        }
    }

    @Override
    public void stop() {
        ConcurrentHashMap<String, Consumer> concurrentHashMap = onsManager.getConsumerConcurrentHashMap();
        concurrentHashMap.forEach((k, v) -> v.shutdown());
    }

    @Override
    public boolean isRunning() {
        return false;
    }

    @Override
    public int getPhase() {
        return 1;
    }


    /**
     * 消息通用处理
     */
    private static class OnsMessageListener implements MessageListener {

        ConsumerAction consumerAction;

        public OnsMessageListener(ConsumerAction consumerAction) {
            this.consumerAction = consumerAction;
        }

        @Override
        public Action consume(Message message, ConsumeContext consumeContext) {
            int reConsumeTimes = message.getReconsumeTimes();
            if (reConsumeTimes >= ONS_MAX_RETRY_TIMES) {
                log.warn("消息重复消费次数达到上限，将被丢弃,消息：{}", message);
                return Action.CommitMessage;
            }

            String messageBody = null;
            try {
                messageBody = new String(message.getBody());
                consumerAction.consume(message.getTag(), messageBody);
            } catch (HuihuaException e) {
                log.error("mq消息消费异常,消息：{},HuihuaException：{}", messageBody, e);
                if (INVALID_PARAMS.getCode().equals(e.getCode())) {
                    //如果是参数错误,重发消息是没用的
                    return Action.CommitMessage;
                } else {
                    return Action.ReconsumeLater;
                }
            } catch (Exception e) {
                log.error("mq消息消费异常,消息：{},Exception：{}", messageBody, e);
                return Action.ReconsumeLater;
            }
            return Action.CommitMessage;
        }
    }


    protected Consumer getConsumer(String topic) {
        ConcurrentHashMap<String, Consumer> concurrentHashMap = onsManager.getConsumerConcurrentHashMap();
        return concurrentHashMap.get(topic);
    }
}
